@@ -1,21 +1,23 @@
 """Serve files directly from the ContentsManager."""
+
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
-import json
+from __future__ import annotations
+
 import mimetypes
 from base64 import decodebytes
-from typing import List
+from typing import Awaitable
 
+from jupyter_core.utils import ensure_async
 from tornado import web
 
-from jupyter_server.auth import authorized
+from jupyter_server.auth.decorator import authorized
 from jupyter_server.base.handlers import JupyterHandler
-from jupyter_server.utils import ensure_async
 
 AUTH_RESOURCE = "contents"
 
 
-class FilesHandler(JupyterHandler):
+class FilesHandler(JupyterHandler, web.StaticFileHandler):
     """serve files via ContentsManager
 
     Normally used when ContentsManager is not a FileContentsManager.
@@ -28,13 +30,15 @@ class FilesHandler(JupyterHandler):
 
     @property
     def content_security_policy(self):
+        """The content security policy."""
         # In case we're serving HTML/SVG, confine any Javascript to a unique
         # origin so it can't interact with the notebook server.
         return super().content_security_policy + "; sandbox allow-scripts"
 
     @web.authenticated
     @authorized
-    def head(self, path):
+    def head(self, path: str) -> Awaitable[None] | None:  # type:ignore[override]
+        """The head response."""
         self.get(path, include_body=False)
         self.check_xsrf_cookie()
         return self.get(path, include_body=False)
@@ -42,11 +46,12 @@ class FilesHandler(JupyterHandler):
     @web.authenticated
     @authorized
     async def get(self, path, include_body=True):
+        """Get a file by path."""
         # /files/ requests must originate from the same site
         self.check_xsrf_cookie()
         cm = self.contents_manager
 
-        if await ensure_async(cm.is_hidden(path)) and not cm.allow_hidden:
+        if not cm.allow_hidden and await ensure_async(cm.is_hidden(path)):
             self.log.info("Refusing to serve hidden file, via 404 Error")
             raise web.HTTPError(404)
 
@@ -75,21 +80,18 @@ class FilesHandler(JupyterHandler):
                 self.set_header("Content-Type", "application/octet-stream")
             elif cur_mime is not None:
                 self.set_header("Content-Type", cur_mime)
+            elif model["format"] == "base64":
+                self.set_header("Content-Type", "application/octet-stream")
             else:
-                if model["format"] == "base64":
-                    self.set_header("Content-Type", "application/octet-stream")
-                else:
-                    self.set_header("Content-Type", "text/plain; charset=UTF-8")
+                self.set_header("Content-Type", "text/plain; charset=UTF-8")
 
         if include_body:
             if model["format"] == "base64":
                 b64_bytes = model["content"].encode("ascii")
                 self.write(decodebytes(b64_bytes))
-            elif model["format"] == "json":
-                self.write(json.dumps(model["content"]))
             else:
                 self.write(model["content"])
             self.flush()
 
 
-default_handlers: List[JupyterHandler] = []
+default_handlers: list[JupyterHandler] = []
