@@ -1,37 +1,47 @@
+"""Kernel gateway managers."""
+
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
+from __future__ import annotations
+
 import asyncio
 import datetime
 import json
 import os
-from logging import Logger
 from queue import Empty, Queue
 from threading import Thread
 from time import monotonic
-from typing import Any, Dict, Optional
+from typing import TYPE_CHECKING, Any, Optional, cast
 
 import websocket
 from jupyter_client.asynchronous.client import AsyncKernelClient
 from jupyter_client.clientabc import KernelClientABC
 from jupyter_client.kernelspec import KernelSpecManager
-from jupyter_client.manager import AsyncKernelManager
 from jupyter_client.managerabc import KernelManagerABC
+from jupyter_core.utils import ensure_async
 from tornado import web
 from tornado.escape import json_decode, json_encode, url_escape, utf8
 from traitlets import DottedObjectName, Instance, Type, default
 
-from .._tz import UTC
-from ..services.kernels.kernelmanager import AsyncMappingKernelManager
+from .._tz import UTC, utcnow
+from ..services.kernels.kernelmanager import (
+    AsyncMappingKernelManager,
+    ServerKernelManager,
+    emit_kernel_action_event,
+)
 from ..services.sessions.sessionmanager import SessionManager
-from ..utils import ensure_async, url_path_join
+from ..utils import url_path_join
 from .gateway_client import GatewayClient, gateway_request
 
+if TYPE_CHECKING:
+    from logging import Logger
+
 
 class GatewayMappingKernelManager(AsyncMappingKernelManager):
     """Kernel manager that supports remote kernels hosted by Jupyter Kernel or Enterprise Gateway."""
 
     # We'll maintain our own set of kernel ids
-    _kernels: Dict[str, "GatewayKernelManager"] = {}
+    _kernels: dict[str, GatewayKernelManager] = {}  # type:ignore[assignment]
 
     @default("kernel_manager_class")
     def _default_kernel_manager_class(self):
@@ -42,9 +52,10 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
         return False  # no need to share zmq contexts
 
     def __init__(self, **kwargs):
+        """Initialize a gateway mapping kernel manager."""
         super().__init__(**kwargs)
         self.kernels_url = url_path_join(
-            GatewayClient.instance().url, GatewayClient.instance().kernels_endpoint
+            GatewayClient.instance().url or "", GatewayClient.instance().kernels_endpoint or ""
         )
 
     def remove_kernel(self, kernel_id):
@@ -54,7 +65,7 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
         except KeyError:
             pass
 
-    async def start_kernel(self, kernel_id=None, path=None, **kwargs):
+    async def start_kernel(self, *, kernel_id=None, path=None, **kwargs):
         """Start a kernel for a session and return its kernel_id.
 
         Parameters
@@ -69,15 +80,13 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
         """
         self.log.info(f"Request start kernel: kernel_id={kernel_id}, path='{path}'")
 
-        if kernel_id is None:
-            if path is not None:
-                kwargs["cwd"] = self.cwd_for_path(path)
+        if kernel_id is None and path is not None:
+            kwargs["cwd"] = self.cwd_for_path(path)
 
         km = self.kernel_manager_factory(parent=self, log=self.log)
         await km.start_kernel(kernel_id=kernel_id, **kwargs)
         kernel_id = km.kernel_id
         self._kernels[kernel_id] = km
-
         # Initialize culling if not already
         if not self._initialized_culler:
             self.initialize_culler()
@@ -95,8 +104,8 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
         """
         model = None
         km = self.get_kernel(str(kernel_id))
-        if km:
-            model = km.kernel
+        if km:  # type:ignore[truthy-bool]
+            model = km.kernel  # type:ignore[attr-defined]
         return model
 
     async def list_kernels(self, **kwargs):
@@ -119,13 +128,36 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
         # Remove any of our kernels that may have been culled on the gateway server
         our_kernels = self._kernels.copy()
         culled_ids = []
-        for kid, _ in our_kernels.items():
+        for kid in our_kernels:
             if kid not in kernel_models:
+                # The upstream kernel was not reported in the list of kernels.
                 self.log.warning(
-                    f"Kernel {kid} no longer active - probably culled on Gateway server."
+                    f"Kernel {kid} not present in the list of kernels - possibly culled on Gateway server."
                 )
-                self._kernels.pop(kid, None)
-                culled_ids.append(kid)  # TODO: Figure out what do with these.
+                try:
+                    # Try to directly refresh the model for this specific kernel in case
+                    # the upstream list of kernels was erroneously incomplete.
+                    #
+                    # That might happen if the case of a proxy that manages multiple
+                    # backends where there could be transient connectivity issues with
+                    # a single backend.
+                    #
+                    # Alternatively, it could happen if there is simply a bug in the
+                    # upstream gateway server.
+                    #
+                    # Either way, including this check improves our reliability in the
+                    # face of such scenarios.
+                    model = await self._kernels[kid].refresh_model()
+                except web.HTTPError:
+                    model = None
+                if model:
+                    kernel_models[kid] = model
+                else:
+                    self.log.warning(
+                        f"Kernel {kid} no longer active - probably culled on Gateway server."
+                    )
+                    self._kernels.pop(kid, None)
+                    culled_ids.append(kid)  # TODO: Figure out what do with these.
         return list(kernel_models.values())
 
     async def shutdown_kernel(self, kernel_id, now=False, restart=False):
@@ -141,7 +173,7 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
             The purpose of this shutdown is to restart the kernel (True)
         """
         km = self.get_kernel(kernel_id)
-        await km.shutdown_kernel(now=now, restart=restart)
+        await ensure_async(km.shutdown_kernel(now=now, restart=restart))
         self.remove_kernel(kernel_id)
 
     async def restart_kernel(self, kernel_id, now=False, **kwargs):
@@ -153,7 +185,7 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
             The id of the kernel to restart.
         """
         km = self.get_kernel(kernel_id)
-        await km.restart_kernel(now=now, **kwargs)
+        await ensure_async(km.restart_kernel(now=now, **kwargs))
 
     async def interrupt_kernel(self, kernel_id, **kwargs):
         """Interrupt a kernel by its kernel uuid.
@@ -164,14 +196,14 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
             The id of the kernel to interrupt.
         """
         km = self.get_kernel(kernel_id)
-        await km.interrupt_kernel()
+        await ensure_async(km.interrupt_kernel())
 
     async def shutdown_all(self, now=False):
         """Shutdown all kernels."""
         kids = list(self._kernels)
         for kernel_id in kids:
             km = self.get_kernel(kernel_id)
-            await km.shutdown_kernel(now=now)
+            await ensure_async(km.shutdown_kernel(now=now))
             self.remove_kernel(kernel_id)
 
     async def cull_kernels(self):
@@ -181,28 +213,56 @@ class GatewayMappingKernelManager(AsyncMappingKernelManager):
 
 
 class GatewayKernelSpecManager(KernelSpecManager):
+    """A gateway kernel spec manager."""
+
     def __init__(self, **kwargs):
+        """Initialize a gateway kernel spec manager."""
         super().__init__(**kwargs)
         base_endpoint = url_path_join(
-            GatewayClient.instance().url, GatewayClient.instance().kernelspecs_endpoint
+            GatewayClient.instance().url or "", GatewayClient.instance().kernelspecs_endpoint
         )
 
         self.base_endpoint = GatewayKernelSpecManager._get_endpoint_for_user_filter(base_endpoint)
         self.base_resource_endpoint = url_path_join(
-            GatewayClient.instance().url,
+            GatewayClient.instance().url or "",
             GatewayClient.instance().kernelspecs_resource_endpoint,
         )
 
     @staticmethod
     def _get_endpoint_for_user_filter(default_endpoint):
+        """Get the endpoint for a user filter."""
         kernel_user = os.environ.get("KERNEL_USERNAME")
         if kernel_user:
-            return "?user=".join([default_endpoint, kernel_user])
+            return f"{default_endpoint}?user={kernel_user}"
         return default_endpoint
 
+    def _replace_path_kernelspec_resources(self, kernel_specs):
+        """Helper method that replaces any gateway base_url with the server's base_url
+        This enables clients to properly route through jupyter_server to a gateway
+        for kernel resources such as logo files
+        """
+        if not self.parent:
+            return {}
+        kernelspecs = kernel_specs["kernelspecs"]
+        for kernel_name in kernelspecs:
+            resources = kernelspecs[kernel_name]["resources"]
+            for resource_name in resources:
+                original_path = resources[resource_name]
+                split_eg_base_url = str.rsplit(original_path, sep="/kernelspecs/", maxsplit=1)
+                if len(split_eg_base_url) > 1:
+                    new_path = url_path_join(
+                        self.parent.base_url, "kernelspecs", split_eg_base_url[1]
+                    )
+                    kernel_specs["kernelspecs"][kernel_name]["resources"][resource_name] = new_path
+                    if original_path != new_path:
+                        self.log.debug(
+                            f"Replaced original kernel resource path {original_path} with new "
+                            f"path {kernel_specs['kernelspecs'][kernel_name]['resources'][resource_name]}"
+                        )
+        return kernel_specs
+
     def _get_kernelspecs_endpoint_url(self, kernel_name=None):
         """Builds a url for the kernels endpoint
-
         Parameters
         ----------
         kernel_name : kernel name (optional)
@@ -213,12 +273,15 @@ class GatewayKernelSpecManager(KernelSpecManager):
         return self.base_endpoint
 
     async def get_all_specs(self):
+        """Get all of the kernel specs for the gateway."""
         fetched_kspecs = await self.list_kernel_specs()
 
         # get the default kernel name and compare to that of this server.
         # If different log a warning and reset the default.  However, the
         # caller of this method will still return this server's value until
         # the next fetch of kernelspecs - at which time they'll match.
+        if not self.parent:
+            return {}
         km = self.parent.kernel_manager
         remote_default_kernel_name = fetched_kspecs.get("default")
         if remote_default_kernel_name != km.default_kernel_name:
@@ -237,6 +300,7 @@ class GatewayKernelSpecManager(KernelSpecManager):
         self.log.debug(f"Request list kernel specs at: {kernel_spec_url}")
         response = await gateway_request(kernel_spec_url, method="GET")
         kernel_specs = json_decode(response.body)
+        kernel_specs = self._replace_path_kernelspec_resources(kernel_specs)
         return kernel_specs
 
     async def get_kernel_spec(self, kernel_name, **kwargs):
@@ -255,12 +319,8 @@ class GatewayKernelSpecManager(KernelSpecManager):
             if error.status_code == 404:
                 # Convert not found to KeyError since that's what the Notebook handler expects
                 # message is not used, but might as well make it useful for troubleshooting
-                raise KeyError(
-                    "kernelspec {kernel_name} not found on Gateway server at: {gateway_url}".format(
-                        kernel_name=kernel_name,
-                        gateway_url=GatewayClient.instance().url,
-                    )
-                ) from error
+                msg = f"kernelspec {kernel_name} not found on Gateway server at: {GatewayClient.instance().url}"
+                raise KeyError(msg) from None
             else:
                 raise
         else:
@@ -295,26 +355,32 @@ class GatewayKernelSpecManager(KernelSpecManager):
 
 
 class GatewaySessionManager(SessionManager):
+    """A gateway session manager."""
+
     kernel_manager = Instance("jupyter_server.gateway.managers.GatewayMappingKernelManager")
 
-    async def kernel_culled(self, kernel_id):
+    async def kernel_culled(self, kernel_id: str) -> bool:  # typing: ignore
         """Checks if the kernel is still considered alive and returns true if it's not found."""
-        kernel = None
+        km: Optional[GatewayKernelManager] = None
         try:
+            # Since we keep the models up-to-date via client polling, use that state to determine
+            # if this kernel no longer exists on the gateway server rather than perform a redundant
+            # fetch operation - especially since this is called at approximately the same interval.
+            # This has the effect of reducing GET /api/kernels requests against the gateway server
+            # by 50%!
+            # Note that should the redundant polling be consolidated, or replaced with an event-based
+            # notification model, this will need to be revisited.
             km = self.kernel_manager.get_kernel(kernel_id)
-            kernel = await km.refresh_model()
-        except Exception:  # Let exceptions here reflect culled kernel
+        except Exception:
+            # Let exceptions here reflect culled kernel
             pass
-        return kernel is None
+        return km is None
 
 
-"""KernelManager class to manage a kernel running on a Gateway Server via the REST API"""
-
-
-class GatewayKernelManager(AsyncKernelManager):
+class GatewayKernelManager(ServerKernelManager):
     """Manages a single kernel remotely via a Gateway Server."""
 
-    kernel_id = None
+    kernel_id: Optional[str] = None  # type:ignore[assignment]
     kernel = None
 
     @default("cache_ports")
@@ -322,14 +388,16 @@ class GatewayKernelManager(AsyncKernelManager):
         return False  # no need to cache ports here
 
     def __init__(self, **kwargs):
+        """Initialize the gateway kernel manager."""
         super().__init__(**kwargs)
         self.kernels_url = url_path_join(
-            GatewayClient.instance().url, GatewayClient.instance().kernels_endpoint
+            GatewayClient.instance().url or "", GatewayClient.instance().kernels_endpoint
         )
         self.kernel_url: str
         self.kernel = self.kernel_id = None
         # simulate busy/activity markers:
-        self.execution_state = self.last_activity = None
+        self.execution_state = "starting"
+        self.last_activity = utcnow()
 
     @property
     def has_kernel(self):
@@ -345,13 +413,13 @@ class GatewayKernelManager(AsyncKernelManager):
 
     def client(self, **kwargs):
         """Create a client configured to connect to our kernel"""
-        kw = {}
+        kw: dict[str, Any] = {}
         kw.update(self.get_connection_info(session=True))
         kw.update(
-            dict(
-                connection_file=self.connection_file,
-                parent=self,
-            )
+            {
+                "connection_file": self.connection_file,
+                "parent": self,
+            }
         )
         kw["kernel_id"] = self.kernel_id
 
@@ -393,7 +461,7 @@ class GatewayKernelManager(AsyncKernelManager):
                 # this kernel manager.  The current kernel manager instance may not have
                 # a parent instance if, say, a server extension is using another application
                 # (e.g., papermill) that uses a KernelManager instance directly.
-                self.parent._kernel_connections[self.kernel_id] = int(model["connections"])
+                self.parent._kernel_connections[self.kernel_id] = int(model["connections"])  # type:ignore[index]
 
         self.kernel = model
         return model
@@ -402,6 +470,9 @@ class GatewayKernelManager(AsyncKernelManager):
     # Kernel management
     # --------------------------------------------------------------------------
 
+    @emit_kernel_action_event(
+        success_msg="Kernel {kernel_id} was started.",
+    )
     async def start_kernel(self, **kwargs):
         """Starts a kernel via HTTP in an asynchronous manner.
 
@@ -419,17 +490,18 @@ class GatewayKernelManager(AsyncKernelManager):
 
             # Let KERNEL_USERNAME take precedent over http_user config option.
             if os.environ.get("KERNEL_USERNAME") is None and GatewayClient.instance().http_user:
-                os.environ["KERNEL_USERNAME"] = GatewayClient.instance().http_user
+                os.environ["KERNEL_USERNAME"] = GatewayClient.instance().http_user or ""
 
+            payload_envs = os.environ.copy()
+            payload_envs.update(kwargs.get("env", {}))  # Add any env entries in this request
+
+            # Build the actual env payload, filtering allowed_envs and those starting with 'KERNEL_'
             kernel_env = {
                 k: v
-                for (k, v) in dict(os.environ).items()
-                if k.startswith("KERNEL_") or k in GatewayClient.instance().env_whitelist.split(",")
+                for (k, v) in payload_envs.items()
+                if k.startswith("KERNEL_") or k in GatewayClient.instance().allowed_envs.split(",")
             }
 
-            # Add any env entries in this request
-            kernel_env.update(kwargs.get("env", {}))
-
             # Convey the full path to where this notebook file is located.
             if kwargs.get("cwd") is not None and kernel_env.get("KERNEL_WORKING_DIR") is None:
                 kernel_env["KERNEL_WORKING_DIR"] = kwargs["cwd"]
@@ -452,6 +524,9 @@ class GatewayKernelManager(AsyncKernelManager):
             self.kernel = await self.refresh_model()
             self.log.info(f"GatewayKernelManager using existing kernel: {self.kernel_id}")
 
+    @emit_kernel_action_event(
+        success_msg="Kernel {kernel_id} was shutdown.",
+    )
     async def shutdown_kernel(self, now=False, restart=False):
         """Attempts to stop the kernel process cleanly via HTTP."""
 
@@ -466,6 +541,9 @@ class GatewayKernelManager(AsyncKernelManager):
                 else:
                     raise
 
+    @emit_kernel_action_event(
+        success_msg="Kernel {kernel_id} was restarted.",
+    )
     async def restart_kernel(self, **kw):
         """Restarts a kernel via HTTP."""
         if self.has_kernel:
@@ -480,6 +558,9 @@ class GatewayKernelManager(AsyncKernelManager):
             )
             self.log.debug("Restart kernel response: %d %s", response.code, response.reason)
 
+    @emit_kernel_action_event(
+        success_msg="Kernel {kernel_id} was interrupted.",
+    )
     async def interrupt_kernel(self):
         """Interrupts the kernel via an HTTP request."""
         if self.has_kernel:
@@ -499,24 +580,27 @@ class GatewayKernelManager(AsyncKernelManager):
         if self.has_kernel:
             # Go ahead and issue a request to get the kernel
             self.kernel = await self.refresh_model()
+            self.log.debug(f"The kernel: {self.kernel} is alive.")
             return True
         else:  # we don't have a kernel
+            self.log.debug(f"The kernel: {self.kernel} no longer exists.")
             return False
 
     def cleanup_resources(self, restart=False):
         """Clean up resources when the kernel is shut down"""
-        pass
 
 
 KernelManagerABC.register(GatewayKernelManager)
 
 
-class ChannelQueue(Queue):
+class ChannelQueue(Queue):  # type:ignore[type-arg]
+    """A queue for a named channel."""
 
     channel_name: Optional[str] = None
     response_router_finished: bool
 
     def __init__(self, channel_name: str, channel_socket: websocket.WebSocket, log: Logger):
+        """Initialize a channel queue."""
         super().__init__()
         self.channel_name = channel_name
         self.channel_socket = channel_socket
@@ -524,10 +608,12 @@ class ChannelQueue(Queue):
         self.response_router_finished = False
 
     async def _async_get(self, timeout=None):
+        """Asynchronously get from the queue."""
         if timeout is None:
             timeout = float("inf")
         elif timeout < 0:
-            raise ValueError("'timeout' must be a non-negative number")
+            msg = "'timeout' must be a non-negative number"
+            raise ValueError(msg)
         end_time = monotonic() + timeout
 
         while True:
@@ -535,12 +621,14 @@ class ChannelQueue(Queue):
                 return self.get(block=False)
             except Empty:
                 if self.response_router_finished:
-                    raise RuntimeError("Response router had finished")
+                    msg = "Response router had finished"
+                    raise RuntimeError(msg) from None
                 if monotonic() > end_time:
                     raise
                 await asyncio.sleep(0)
 
-    async def get_msg(self, *args: Any, **kwargs: Any) -> dict:
+    async def get_msg(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
+        """Get a message from the queue."""
         timeout = kwargs.get("timeout", 1)
         msg = await self._async_get(timeout=timeout)
         self.log.debug(
@@ -549,9 +637,10 @@ class ChannelQueue(Queue):
             )
         )
         self.task_done()
-        return msg
+        return cast("dict[str, Any]", msg)
 
-    def send(self, msg: dict) -> None:
+    def send(self, msg: dict[str, Any]) -> None:
+        """Send a message to the queue."""
         message = json.dumps(msg, default=ChannelQueue.serialize_datetime).replace("</", "<\\/")
         self.log.debug(
             "Sending message on channel: {}, msg_id: {}, msg_type: {}".format(
@@ -562,14 +651,16 @@ class ChannelQueue(Queue):
 
     @staticmethod
     def serialize_datetime(dt):
+        """Serialize a datetime object."""
         if isinstance(dt, datetime.datetime):
             return dt.timestamp()
         return None
 
     def start(self) -> None:
-        pass
+        """Start the queue."""
 
     def stop(self) -> None:
+        """Stop the queue."""
         if not self.empty():
             # If unprocessed messages are detected, drain the queue collecting non-status
             # messages.  If any remain that are not 'shutdown_reply' and this is not iopub
@@ -583,17 +674,19 @@ class ChannelQueue(Queue):
                 return
             if len(msgs):
                 self.log.warning(
-                    "Stopping channel '{}' with {} unprocessed non-status messages: {}.".format(
-                        self.channel_name, len(msgs), msgs
-                    )
+                    f"Stopping channel '{self.channel_name}' with {len(msgs)} unprocessed non-status messages: {msgs}."
                 )
 
     def is_alive(self) -> bool:
+        """Whether the queue is alive."""
         return self.channel_socket is not None
 
 
 class HBChannelQueue(ChannelQueue):
+    """A queue for the heartbeat channel."""
+
     def is_beating(self) -> bool:
+        """Whether the channel is beating."""
         # Just use the is_alive status for now
         return self.is_alive()
 
@@ -618,14 +711,15 @@ class GatewayKernelClient(AsyncKernelClient):
     # flag for whether execute requests should be allowed to call raw_input:
     allow_stdin = False
     _channels_stopped: bool
-    _channel_queues: Optional[Dict[str, ChannelQueue]]
-    _control_channel: Optional[ChannelQueue]
-    _hb_channel: Optional[ChannelQueue]
-    _stdin_channel: Optional[ChannelQueue]
-    _iopub_channel: Optional[ChannelQueue]
-    _shell_channel: Optional[ChannelQueue]
+    _channel_queues: Optional[dict[str, ChannelQueue]]
+    _control_channel: Optional[ChannelQueue]  # type:ignore[assignment]
+    _hb_channel: Optional[ChannelQueue]  # type:ignore[assignment]
+    _stdin_channel: Optional[ChannelQueue]  # type:ignore[assignment]
+    _iopub_channel: Optional[ChannelQueue]  # type:ignore[assignment]
+    _shell_channel: Optional[ChannelQueue]  # type:ignore[assignment]
 
     def __init__(self, kernel_id, **kwargs):
+        """Initialize a gateway kernel client."""
         super().__init__(**kwargs)
         self.kernel_id = kernel_id
         self.channel_socket: Optional[websocket.WebSocket] = None
@@ -646,7 +740,7 @@ class GatewayKernelClient(AsyncKernelClient):
         """
 
         ws_url = url_path_join(
-            GatewayClient.instance().ws_url,
+            GatewayClient.instance().ws_url or "",
             GatewayClient.instance().kernels_endpoint,
             url_escape(self.kernel_id),
             "channels",
