@@ -2,20 +2,25 @@
 
 Preliminary documentation at https://github.com/ipython/ipython/wiki/IPEP-25%3A-Registry-of-installed-kernels#rest-api
 """
+
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
+from __future__ import annotations
+
 import glob
 import json
 import os
+from typing import Any
 
 pjoin = os.path.join
 
+from jupyter_core.utils import ensure_async
 from tornado import web
 
-from jupyter_server.auth import authorized
+from jupyter_server.auth.decorator import authorized
 
 from ...base.handlers import APIHandler
-from ...utils import ensure_async, url_path_join, url_unescape
+from ...utils import url_path_join, url_unescape
 
 AUTH_RESOURCE = "kernelspecs"
 
@@ -25,7 +30,6 @@ def kernelspec_model(handler, name, spec_dict, resource_dir):
     d = {"name": name, "spec": spec_dict, "resources": {}}
 
     # Add resource files if they exist
-    resource_dir = resource_dir
     for resource in ["kernel.js", "kernel.css"]:
         if os.path.exists(pjoin(resource_dir, resource)):
             d["resources"][resource] = url_path_join(
@@ -49,16 +53,21 @@ def is_kernelspec_model(spec_dict):
 
 
 class KernelSpecsAPIHandler(APIHandler):
+    """A kernel spec API handler."""
+
     auth_resource = AUTH_RESOURCE
 
 
 class MainKernelSpecHandler(KernelSpecsAPIHandler):
+    """The root kernel spec handler."""
+
     @web.authenticated
     @authorized
     async def get(self):
+        """Get the list of kernel specs."""
         ksm = self.kernel_spec_manager
         km = self.kernel_manager
-        model = {}
+        model: dict[str, Any] = {}
         model["default"] = km.default_kernel_name
         model["kernelspecs"] = specs = {}
         kspecs = await ensure_async(ksm.get_all_specs())
@@ -82,9 +91,12 @@ class MainKernelSpecHandler(KernelSpecsAPIHandler):
 
 
 class KernelSpecHandler(KernelSpecsAPIHandler):
+    """A handler for an individual kernel spec."""
+
     @web.authenticated
     @authorized
     async def get(self, kernel_name):
+        """Get a kernel spec model."""
         ksm = self.kernel_spec_manager
         kernel_name = url_unescape(kernel_name)
         try:
