@@ -1,6 +1,7 @@
 """A mock extension module with a list of extensions
 to load in various tests.
 """
+
 from .app import MockExtensionApp
 
 
@@ -9,16 +10,10 @@ from .app import MockExtensionApp
 def _jupyter_server_extension_points():
     return [
         {
-            'module': 'tests.extension.mockextensions.app',
-            'app': MockExtensionApp
+            "module": "tests.extension.mockextensions.app",
+            "app": MockExtensionApp,
         },
-        {
-            'module': 'tests.extension.mockextensions.mock1'
-        },
-        {
-            'module': 'tests.extension.mockextensions.mock2'
-        },
-        {
-            'module': 'tests.extension.mockextensions.mock3'
-        }
+        {"module": "tests.extension.mockextensions.mock1"},
+        {"module": "tests.extension.mockextensions.mock2"},
+        {"module": "tests.extension.mockextensions.mock3"},
     ]
