@@ -1,16 +1,40 @@
 import getpass
+import json
 import logging
 import os
 import pathlib
+import sys
+import warnings
 from unittest.mock import patch
 
 import pytest
 from jupyter_core.application import NoStart
+from tornado import web
 from traitlets import TraitError
+from traitlets.config import Config
 from traitlets.tests.utils import check_help_all_output
 
+from jupyter_server.auth.decorator import allow_unauthenticated, authorized
 from jupyter_server.auth.security import passwd_check
-from jupyter_server.serverapp import JupyterPasswordApp, ServerApp, list_running_servers
+from jupyter_server.serverapp import (
+    JupyterPasswordApp,
+    JupyterServerListApp,
+    ServerApp,
+    ServerWebApplication,
+    _has_tornado_web_authenticated,
+    list_running_servers,
+    random_ports,
+)
+from jupyter_server.services.contents.filemanager import (
+    AsyncFileContentsManager,
+    FileContentsManager,
+)
+from jupyter_server.utils import pathname2url, urljoin
+
+
+@pytest.fixture(params=[FileContentsManager, AsyncFileContentsManager])
+def jp_file_contents_manager_class(request, tmp_path):
+    return request.param
 
 
 def test_help_output():
@@ -18,6 +42,49 @@ def test_help_output():
     check_help_all_output("jupyter_server")
 
 
+@pytest.mark.parametrize(
+    "format",
+    [
+        "json",
+        "jsonlist",
+        "",
+    ],
+)
+def test_server_list(jp_configurable_serverapp, capsys, format):
+    app = jp_configurable_serverapp(log=logging.getLogger())
+
+    app.write_server_info_file()
+
+    capsys.readouterr()
+    listapp = JupyterServerListApp(
+        parent=app,
+    )
+    if format:
+        setattr(listapp, format, True)
+    listapp.start()
+    captured = capsys.readouterr()
+    sys.stdout.write(captured.out)
+    sys.stderr.write(captured.err)
+    out = captured.out.strip()
+
+    if not format:
+        assert "Currently running servers:" in out
+        assert app.connection_url in out
+        assert len(out.splitlines()) == 2
+        return
+
+    if format == "jsonlist":
+        servers = json.loads(out)
+    elif format == "json":
+        servers = [json.loads(line) for line in out.splitlines()]
+    assert len(servers) == 1
+    sinfo = servers[0]
+
+    assert sinfo["port"] == app.port
+    assert sinfo["url"] == app.connection_url
+    assert sinfo["version"] == app.version
+
+
 def test_server_info_file(tmp_path, jp_configurable_serverapp):
     app = jp_configurable_serverapp(log=logging.getLogger())
 
@@ -34,7 +101,7 @@ def test_server_info_file(tmp_path, jp_configurable_serverapp):
     app.remove_server_info_file()
 
     assert list(list_running_servers(app.runtime_dir)) == []
-    app.remove_server_info_file
+    app.remove_server_info_file()
 
 
 def test_root_dir(tmp_path, jp_configurable_serverapp):
@@ -77,7 +144,7 @@ def test_valid_root_dir(valid_root_dir, jp_configurable_serverapp):
     assert app.root_dir == root_dir
 
 
-def test_generate_config(tmp_path, jp_configurable_serverapp):
+async def test_generate_config(tmp_path, jp_configurable_serverapp):
     app = jp_configurable_serverapp(config_dir=str(tmp_path))
     app.initialize(["--generate-config", "--allow-root"])
     with pytest.raises(NoStart):
@@ -95,8 +162,28 @@ def test_server_password(tmp_path, jp_configurable_serverapp):
         app.start()
         sv = jp_configurable_serverapp()
         sv.load_config_file()
-        assert sv.password != ""
-        passwd_check(sv.password, password)
+        assert sv.identity_provider.hashed_password != ""
+        passwd_check(sv.identity_provider.hashed_password, password)
+
+
+@pytest.mark.parametrize(
+    "env,expected",
+    [
+        ["yes", True],
+        ["Yes", True],
+        ["True", True],
+        ["true", True],
+        ["TRUE", True],
+        ["no", False],
+        ["nooo", False],
+        ["FALSE", False],
+        ["false", False],
+    ],
+)
+def test_allow_unauthenticated_env_var(jp_configurable_serverapp, env, expected):
+    with patch.dict("os.environ", {"JUPYTER_SERVER_ALLOW_UNAUTHENTICATED_ACCESS": env}):
+        app = jp_configurable_serverapp()
+        assert app.allow_unauthenticated_access == expected
 
 
 def test_list_running_servers(jp_serverapp, jp_web_app):
@@ -239,12 +326,14 @@ def test_urls(config, public_url, local_url, connection_url):
     # Verify we're working with a clean instance.
     ServerApp.clear_instance()
     serverapp = ServerApp.instance(**config)
+    serverapp.init_configurables()
+    token = serverapp.identity_provider.token
     # If a token is generated (not set by config), update
     # expected_url with token.
-    if serverapp._token_generated:
-        public_url = public_url.replace("<generated>", serverapp.token)
-        local_url = local_url.replace("<generated>", serverapp.token)
-        connection_url = connection_url.replace("<generated>", serverapp.token)
+    if serverapp.identity_provider.token_generated:
+        public_url = public_url.replace("<generated>", token)
+        local_url = local_url.replace("<generated>", token)
+        connection_url = connection_url.replace("<generated>", token)
     assert serverapp.public_url == public_url
     assert serverapp.local_url == local_url
     assert serverapp.connection_url == connection_url
@@ -254,14 +343,18 @@ def test_urls(config, public_url, local_url, connection_url):
 
 # Preferred dir tests
 # ----------------------------------------------------------------------------
+@pytest.mark.filterwarnings("ignore::FutureWarning")
 def test_valid_preferred_dir(tmp_path, jp_configurable_serverapp):
     path = str(tmp_path)
     app = jp_configurable_serverapp(root_dir=path, preferred_dir=path)
     assert app.root_dir == path
     assert app.preferred_dir == path
     assert app.root_dir == app.preferred_dir
+    assert app.contents_manager.root_dir == path
+    assert app.contents_manager.preferred_dir == ""
 
 
+@pytest.mark.filterwarnings("ignore::FutureWarning")
 def test_valid_preferred_dir_is_root_subdir(tmp_path, jp_configurable_serverapp):
     path = str(tmp_path)
     path_subdir = str(tmp_path / "subdir")
@@ -270,37 +363,75 @@ def test_valid_preferred_dir_is_root_subdir(tmp_path, jp_configurable_serverapp)
     assert app.root_dir == path
     assert app.preferred_dir == path_subdir
     assert app.preferred_dir.startswith(app.root_dir)
+    assert app.contents_manager.preferred_dir == "subdir"
 
 
 def test_valid_preferred_dir_does_not_exist(tmp_path, jp_configurable_serverapp):
     path = str(tmp_path)
     path_subdir = str(tmp_path / "subdir")
     with pytest.raises(TraitError) as error:
-        app = jp_configurable_serverapp(root_dir=path, preferred_dir=path_subdir)
+        jp_configurable_serverapp(root_dir=path, preferred_dir=path_subdir)
 
     assert "No such preferred dir:" in str(error)
 
 
+@pytest.mark.filterwarnings("ignore::DeprecationWarning")
+def test_preferred_dir_validation_sync_regression(
+    tmp_path, jp_configurable_serverapp, jp_file_contents_manager_class
+):
+    path = str(tmp_path)
+    path_subdir = str(tmp_path / "subdir")
+    os.makedirs(path_subdir, exist_ok=True)
+    app = jp_configurable_serverapp(
+        root_dir=path,
+        contents_manager_class=jp_file_contents_manager_class,
+    )
+    app.contents_manager.preferred_dir = path_subdir
+    assert app.preferred_dir == path_subdir
+    assert app.preferred_dir.startswith(app.root_dir)
+    assert app.contents_manager.preferred_dir == "subdir"
+
+
+# This tests some deprecated behavior as well
+@pytest.mark.filterwarnings("ignore::FutureWarning")
 @pytest.mark.parametrize(
-    "root_dir_loc,preferred_dir_loc",
+    "root_dir_loc,preferred_dir_loc,config_target",
     [
-        ("cli", "cli"),
-        ("cli", "config"),
-        ("cli", "default"),
-        ("config", "cli"),
-        ("config", "config"),
-        ("config", "default"),
-        ("default", "cli"),
-        ("default", "config"),
-        ("default", "default"),
+        ("cli", "cli", "ServerApp"),
+        ("cli", "cli", "FileContentsManager"),
+        ("cli", "config", "ServerApp"),
+        ("cli", "config", "FileContentsManager"),
+        ("cli", "default", "ServerApp"),
+        ("cli", "default", "FileContentsManager"),
+        ("config", "cli", "ServerApp"),
+        ("config", "cli", "FileContentsManager"),
+        ("config", "config", "ServerApp"),
+        ("config", "config", "FileContentsManager"),
+        ("config", "default", "ServerApp"),
+        ("config", "default", "FileContentsManager"),
+        ("default", "cli", "ServerApp"),
+        ("default", "cli", "FileContentsManager"),
+        ("default", "config", "ServerApp"),
+        ("default", "config", "FileContentsManager"),
+        ("default", "default", "ServerApp"),
+        ("default", "default", "FileContentsManager"),
     ],
 )
 def test_preferred_dir_validation(
-    root_dir_loc, preferred_dir_loc, tmp_path, jp_config_dir, jp_configurable_serverapp
+    root_dir_loc,
+    preferred_dir_loc,
+    config_target,
+    tmp_path,
+    jp_config_dir,
+    jp_configurable_serverapp,
 ):
     expected_root_dir = str(tmp_path)
-    expected_preferred_dir = str(tmp_path / "subdir")
-    os.makedirs(expected_preferred_dir, exist_ok=True)
+
+    os_preferred_dir = str(tmp_path / "subdir")
+    os.makedirs(os_preferred_dir, exist_ok=True)
+    config_preferred_dir = os_preferred_dir if config_target == "ServerApp" else "subdir"
+    config_preferred_dir = config_preferred_dir + "/"  # add trailing slash to ensure it is removed
+    expected_preferred_dir = "subdir"
 
     argv = []
     kwargs = {"root_dir": None}
@@ -311,33 +442,33 @@ def test_preferred_dir_validation(
         config_file = jp_config_dir.joinpath("jupyter_server_config.py")
 
     if root_dir_loc == "cli":
-        argv.append(f"--ServerApp.root_dir={expected_root_dir}")
+        argv.append(f"--{config_target}.root_dir={expected_root_dir}")
     if root_dir_loc == "config":
-        config_lines.append(f'c.ServerApp.root_dir = r"{expected_root_dir}"')
+        config_lines.append(f'c.{config_target}.root_dir = r"{expected_root_dir}"')
     if root_dir_loc == "default":
         expected_root_dir = os.getcwd()
 
     if preferred_dir_loc == "cli":
-        argv.append(f"--ServerApp.preferred_dir={expected_preferred_dir}")
+        argv.append(f"--{config_target}.preferred_dir={config_preferred_dir}")
     if preferred_dir_loc == "config":
-        config_lines.append(f'c.ServerApp.preferred_dir = r"{expected_preferred_dir}"')
+        config_lines.append(f'c.{config_target}.preferred_dir = r"{config_preferred_dir}"')
     if preferred_dir_loc == "default":
-        expected_preferred_dir = expected_root_dir
+        expected_preferred_dir = ""
 
     if config_file is not None:
         config_file.write_text("\n".join(config_lines))
 
     if argv:
-        kwargs["argv"] = argv  # type:ignore
+        kwargs["argv"] = argv  # type:ignore[assignment]
 
     if root_dir_loc == "default" and preferred_dir_loc != "default":  # error expected
         with pytest.raises(SystemExit):
             jp_configurable_serverapp(**kwargs)
     else:
         app = jp_configurable_serverapp(**kwargs)
-        assert app.root_dir == expected_root_dir
-        assert app.preferred_dir == expected_preferred_dir
-        assert app.preferred_dir.startswith(app.root_dir)
+        assert app.contents_manager.root_dir == expected_root_dir
+        assert app.contents_manager.preferred_dir == expected_preferred_dir
+        assert ".." not in expected_preferred_dir
 
 
 def test_invalid_preferred_dir_does_not_exist(tmp_path, jp_configurable_serverapp):
@@ -360,42 +491,171 @@ def test_invalid_preferred_dir_does_not_exist_set(tmp_path, jp_configurable_serv
     assert "No such preferred dir:" in str(error)
 
 
+@pytest.mark.filterwarnings("ignore::FutureWarning")
 def test_invalid_preferred_dir_not_root_subdir(tmp_path, jp_configurable_serverapp):
     path = str(tmp_path / "subdir")
     os.makedirs(path, exist_ok=True)
     not_subdir_path = str(tmp_path)
 
+    with pytest.raises(SystemExit):
+        jp_configurable_serverapp(root_dir=path, preferred_dir=not_subdir_path)
+
+
+async def test_invalid_preferred_dir_not_root_subdir_set(tmp_path, jp_configurable_serverapp):
+    path = str(tmp_path / "subdir")
+    os.makedirs(path, exist_ok=True)
+    not_subdir_path = os.path.relpath(tmp_path, path)
+
+    app = jp_configurable_serverapp(root_dir=path)
     with pytest.raises(TraitError) as error:
-        app = jp_configurable_serverapp(root_dir=path, preferred_dir=not_subdir_path)
+        app.contents_manager.preferred_dir = not_subdir_path
 
-    assert "preferred_dir must be equal or a subdir of root_dir. " in str(error)
+    assert "is outside root contents directory" in str(error.value)
 
 
-def test_invalid_preferred_dir_not_root_subdir_set(tmp_path, jp_configurable_serverapp):
+async def test_absolute_preferred_dir_not_root_subdir_set(tmp_path, jp_configurable_serverapp):
     path = str(tmp_path / "subdir")
     os.makedirs(path, exist_ok=True)
     not_subdir_path = str(tmp_path)
 
     app = jp_configurable_serverapp(root_dir=path)
+
     with pytest.raises(TraitError) as error:
-        app.preferred_dir = not_subdir_path
+        app.contents_manager.preferred_dir = not_subdir_path
 
-    assert "preferred_dir must be equal or a subdir of root_dir. " in str(error)
+    if os.name == "nt":
+        assert "is not a relative API path" in str(error.value)
+    else:
+        assert "Preferred directory not found" in str(error.value)
 
 
-def test_observed_root_dir_updates_preferred_dir(tmp_path, jp_configurable_serverapp):
-    path = str(tmp_path)
-    new_path = str(tmp_path / "subdir")
-    os.makedirs(new_path, exist_ok=True)
-
-    app = jp_configurable_serverapp(root_dir=path, preferred_dir=path)
-    app.root_dir = new_path
-    assert app.preferred_dir == new_path
+def test_random_ports():
+    ports = list(random_ports(500, 50))
+    assert len(ports) == 50
 
 
-def test_observed_root_dir_does_not_update_preferred_dir(tmp_path, jp_configurable_serverapp):
-    path = str(tmp_path)
-    new_path = str(tmp_path.parent)
-    app = jp_configurable_serverapp(root_dir=path, preferred_dir=path)
-    app.root_dir = new_path
-    assert app.preferred_dir == path
+def test_server_web_application(jp_serverapp):
+    server: ServerApp = jp_serverapp
+    server.default_url = "/foo"
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore")
+        app = ServerWebApplication(
+            server,
+            [],
+            server.kernel_manager,
+            server.contents_manager,
+            server.session_manager,
+            server.kernel_manager,
+            server.config_manager,
+            server.event_logger,
+            [],
+            server.log,
+            server.base_url,
+            server.default_url,
+            {},
+            {},
+        )
+    app.init_handlers([], app.settings)
+
+
+def test_misc(jp_serverapp, tmp_path):
+    app: ServerApp = jp_serverapp
+    assert app.terminals_enabled is True
+    app.extra_args = [str(tmp_path)]
+    app.parse_command_line([])
+
+
+def test_deprecated_props(jp_serverapp, tmp_path):
+    app: ServerApp = jp_serverapp
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore")
+        app.cookie_options = dict(foo=1)
+        app.get_secure_cookie_kwargs = dict(bar=1)
+        app.notebook_dir = str(tmp_path)
+        app.server_extensions = dict(foo=True)
+        app.kernel_ws_protocol = "foo"
+        app.limit_rate = True
+        app.iopub_msg_rate_limit = 10
+        app.iopub_data_rate_limit = 10
+        app.rate_limit_window = 10
+    with pytest.raises(SystemExit):
+        app.pylab = "foo"
+
+
+def test_signals(jp_serverapp):
+    app: ServerApp = jp_serverapp
+    app.answer_yes = True
+    app._restore_sigint_handler()
+    app._handle_sigint(None, None)
+    app._confirm_exit()
+    app._signal_info(None, None)
+
+
+async def test_shutdown_no_activity(jp_serverapp):
+    app: ServerApp = jp_serverapp
+    app.extension_manager.extensions = {}
+    app.exit = lambda _: None  # type:ignore[assignment,misc]
+    app.shutdown_no_activity()
+    app.shutdown_no_activity_timeout = 1
+    app.init_shutdown_no_activity()
+
+
+def test_running_server_info(jp_serverapp):
+    app: ServerApp = jp_serverapp
+    app.running_server_info(True)
+
+
+@pytest.mark.parametrize("should_exist", [True, False])
+def test_browser_open_files(jp_configurable_serverapp, should_exist, caplog):
+    app = jp_configurable_serverapp(no_browser_open_file=not should_exist)
+    assert os.path.exists(app.browser_open_file) == should_exist
+    url = urljoin("file:", pathname2url(app.browser_open_file))
+    url_messages = [rec.message for rec in caplog.records if url in rec.message]
+    assert url_messages if should_exist else not url_messages
+
+
+def test_deprecated_notebook_dir_priority(jp_configurable_serverapp, tmp_path):
+    notebook_dir = tmp_path / "notebook"
+    notebook_dir.mkdir()
+    cli_dir = tmp_path / "cli"
+    cli_dir.mkdir()
+
+    app = jp_configurable_serverapp(argv=[str(cli_dir)], root_dir=None)
+    assert app._root_dir_set
+
+    # simulate delayed loading of notebook_dir config
+    # this should _not_ take priority over an explicitly set root_dir
+    # as done by notebook_shim
+    cfg = Config()
+    cfg.ServerApp.notebook_dir = str(notebook_dir)
+    app.update_config(cfg)
+    assert app.root_dir == str(cli_dir)
+
+
+def test_immutable_cache_trait():
+    # Verify we're working with a clean instance.
+    ServerApp.clear_instance()
+    kwargs = {"static_immutable_cache": "/test/immutable"}
+    serverapp = ServerApp.instance(**kwargs)
+    serverapp.init_configurables()
+    serverapp.init_webapp()
+    assert serverapp.web_app.settings["static_immutable_cache"] == ["/test/immutable"]
+
+
+def test():
+    pass
+
+
+@pytest.mark.parametrize(
+    "method, expected",
+    [
+        [test, False],
+        [allow_unauthenticated(test), False],
+        [authorized(test), False],
+        [web.authenticated(test), True],
+        [web.authenticated(authorized(test)), True],
+        [authorized(web.authenticated(test)), False],  # wrong order!
+    ],
+)
+def test_tornado_authentication_detection(method, expected):
+    assert _has_tornado_web_authenticated(method) == expected
